"""
Computes the tendencies to qt and θ_liq_ice due to precipitation formation
(autoconversion + accretion)
"""
function precipitation_formation(
    param_set::APS,
    precip_model::AbstractPrecipitationModel,
    prog_gm,
    area::FT,
    z::FT,
    Δt::Real,
    ts,
) where {FT}
    thermo_params = TCP.thermodynamics_params(param_set)

    microphys_params = TCP.microphysics_params(param_set)
    # TODO - when using adaptive timestepping we are limiting the source terms
    #        with the previous timestep Δt
    qt_tendency = FT(0)
    ql_tendency = FT(0)
    qi_tendency = FT(0)
    qr_tendency = FT(0)
    qs_tendency = FT(0)
    θ_liq_ice_tendency = FT(0)
    e_tot_tendency = FT(0)

    if area > 0

        q = TD.PhasePartition(thermo_params, ts)

        Π_m = TD.exner(thermo_params, ts)
        c_pm = TD.cp_m(thermo_params, ts)
        L_v0 = TCP.LH_v0(param_set)
        L_s0 = TCP.LH_s0(param_set)
        I_l = TD.internal_energy_liquid(thermo_params, ts)
        I_i = TD.internal_energy_ice(thermo_params, ts)
        I = TD.internal_energy(thermo_params, ts)
        Φ = geopotential(thermo_params, z)

        if precip_model isa Microphysics0Moment
            qsat = TD.q_vap_saturation(thermo_params, ts)
            λ = TD.liquid_fraction(thermo_params, ts)

            S_qt =
                -min(
                    (q.liq + q.ice) / Δt,
                    -CM0.remove_precipitation(microphys_params, q, qsat),
                )

            qr_tendency -= S_qt * λ
            qs_tendency -= S_qt * (1 - λ)
            qt_tendency += S_qt
            ql_tendency += S_qt * λ
            qi_tendency += S_qt * (1 - λ)
            θ_liq_ice_tendency -=
                S_qt / Π_m / c_pm * (L_v0 * λ + L_s0 * (1 - λ))
            e_tot_tendency += (λ * I_l + (1 - λ) * I_i + Φ) * S_qt
        end

        if precip_model isa Microphysics1Moment
            T = TD.air_temperature(thermo_params, ts)
            T_fr = TCP.T_freeze(param_set)
            c_vl = TCP.cv_l(param_set)
            c_vm = TD.cv_m(thermo_params, ts)
            Rm = TD.gas_constant_air(thermo_params, ts)
            Lf = TD.latent_heat_fusion(thermo_params, ts)

            ρ = prog_gm.ρ
            qr = max(FT(0), prog_gm.ρq_rai) / ρ
            qs = max(FT(0), prog_gm.ρq_sno) / ρ

            # Autoconversion of cloud ice to snow is done with a simplified rate.
            # The saturation adjustment scheme prevents using the
            # 1-moment snow autoconversion rate that assumes
            # that the supersaturation is present in the domain.
            S_qt_rain =
                -min(
                    q.liq / Δt,
                    CM1.conv_q_liq_to_q_rai(microphys_params, q.liq),
                )
            S_qt_snow =
                -min(
                    q.ice / Δt,
                    CM1.conv_q_ice_to_q_sno_no_supersat(
                        microphys_params,
                        q.ice,
                    ),
                )
            qr_tendency -= S_qt_rain
            qs_tendency -= S_qt_snow
            qt_tendency += S_qt_rain + S_qt_snow
            ql_tendency += S_qt_rain
            qi_tendency += S_qt_snow
            θ_liq_ice_tendency -=
                1 / Π_m / c_pm * (L_v0 * S_qt_rain + L_s0 * S_qt_snow)
            e_tot_tendency += S_qt_rain * (I_l + Φ) + S_qt_snow * (I_i + Φ)

            # accretion cloud water + rain
            S_qr = min(
                q.liq / Δt,
                CM1.accretion(
                    microphys_params,
                    liq_type,
                    rain_type,
                    q.liq,
                    qr,
                    ρ,
                ),
            )
            qr_tendency += S_qr
            qt_tendency -= S_qr
            ql_tendency -= S_qr
            θ_liq_ice_tendency += S_qr / Π_m / c_pm * L_v0
            e_tot_tendency -= S_qr * (I_l + Φ)

            # accretion cloud ice + snow
            S_qs = min(
                q.ice / Δt,
                CM1.accretion(
                    microphys_params,
                    ice_type,
                    snow_type,
                    q.ice,
                    qs,
                    ρ,
                ),
            )
            qs_tendency += S_qs
            qt_tendency -= S_qs
            qi_tendency -= S_qs
            θ_liq_ice_tendency += S_qs / Π_m / c_pm * L_s0
            e_tot_tendency -= S_qs * (I_i + Φ)

            # sink of cloud water via accretion cloud water + snow
            S_qt =
                -min(
                    q.liq / Δt,
                    CM1.accretion(
                        microphys_params,
                        liq_type,
                        snow_type,
                        q.liq,
                        qs,
                        ρ,
                    ),
                )
            if T < T_fr # cloud droplets freeze to become snow)
                qs_tendency -= S_qt
                qt_tendency += S_qt
                ql_tendency += S_qt
                θ_liq_ice_tendency -= S_qt / Π_m / c_pm * Lf * (1 + Rm / c_vm)
                e_tot_tendency += S_qt * (I_i + Φ)
            else # snow melts, both cloud water and snow become rain
                α::FT = c_vl / Lf * (T - T_fr)
                qt_tendency += S_qt
                ql_tendency += S_qt
                qs_tendency += S_qt * α
                qr_tendency -= S_qt * (1 + α)
                θ_liq_ice_tendency +=
                    S_qt / Π_m / c_pm * (Lf * (1 + Rm / c_vm) * α - L_v0)
                e_tot_tendency += S_qt * ((1 + α) * I_l - α * I_i + Φ)
            end

            # sink of cloud ice via accretion cloud ice - rain
            S_qt =
                -min(
                    q.ice / Δt,
                    CM1.accretion(
                        microphys_params,
                        ice_type,
                        rain_type,
                        q.ice,
                        qr,
                        ρ,
                    ),
                )
            # sink of rain via accretion cloud ice - rain
            S_qr =
                -min(
                    qr / Δt,
                    CM1.accretion_rain_sink(microphys_params, q.ice, qr, ρ),
                )
            qt_tendency += S_qt
            qi_tendency += S_qt
            qr_tendency += S_qr
            qs_tendency += -(S_qt + S_qr)
            θ_liq_ice_tendency -=
                1 / Π_m / c_pm * (S_qr * Lf * (1 + Rm / c_vm) + S_qt * L_s0)
            e_tot_tendency += S_qt * (I_i + Φ)
            e_tot_tendency -= S_qr * Lf

            # accretion rain - snow
            if T < T_fr
                S_qs = min(
                    qr / Δt,
                    CM1.accretion_snow_rain(
                        microphys_params,
                        snow_type,
                        rain_type,
                        qs,
                        qr,
                        ρ,
                    ),
                )
            else
                S_qs =
                    -min(
                        qs / Δt,
                        CM1.accretion_snow_rain(
                            microphys_params,
                            rain_type,
                            snow_type,
                            qr,
                            qs,
                            ρ,
                        ),
                    )
            end
            qs_tendency += S_qs
            qr_tendency -= S_qs
            θ_liq_ice_tendency += S_qs * Lf / Π_m / c_vm
            e_tot_tendency += S_qs * Lf
        end
    end
    return PrecipFormation{FT}(
        θ_liq_ice_tendency,
        e_tot_tendency,
        qt_tendency,
        ql_tendency,
        qi_tendency,
        qr_tendency,
        qs_tendency,
    )
end
